from typing import Optional

from ..base.common import ACROSSBase
from .schema import ResolveGetSchema, ResolveSchema


class Resolve(ACROSSBase):
    """
    Represents a resolver for resolving celestial object coordinates.

    Attributes
    ----------
    name : Optional[str]
        The name of the celestial object.
    ra : Optional[float]
        The right ascension of the celestial object.
    dec : Optional[float]
        The declination of the celestial object.
    resolver : Optional[str]
        The resolver used for resolving the coordinates.
    """

    # Type hints
    name: Optional[str] = None
    ra: Optional[float] = None
    dec: Optional[float] = None
    resolver: Optional[str] = None

    _mission = "ACROSS"
    _api_name = "Resolve"
    _schema = ResolveSchema
    _get_schema = ResolveGetSchema

    def __init__(self, name: Optional[str] = None, **kwargs):
        self.name = name
        for k, a in kwargs.items():
            setattr(self, k, a)


class ACROSSResolveName:
    """_summary_

    Returns
    -------
    _type_
        _description_
    """

    ra: Optional[float]
    dec: Optional[float]
    _name: str

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, targname: str):
        """Set name

        Parameters
        ----------
        targname : str
            Target name that can be resolved by the Resolve class
        """
        self._name = targname
        if hasattr(self, "ra") is False or self.ra is None:
            r = Resolve(name=targname)
            r.get()
            self.ra = r.ra
            self.dec = r.dec


def resolve(name: str) -> Resolve:
    """Resolve a target name to coordinates

    Parameters
    ----------
    name : str
        Target name that can be resolved by the Resolve class

    Returns
    -------
    Resolve
        A Resolve object with the coordinates of the target
    """
    r = Resolve(name=name)
    r.get()
    return r



from datetime import datetime
from typing import Optional

from pydantic import model_validator

from ..base.schema import BaseSchema


class HelloSchema(BaseSchema):
    """
    Schema defining the returned attributes of the  ACROSS API Hello class.
    """

    hello: str


class HelloGetSchema(BaseSchema):
    """
    Schema to validate input parameters of ACROSS API Hello class.
    """

    name: Optional[str] = None


class ResolveSchema(BaseSchema):
    """
    A schema for resolving astronomical coordinates.
    """

    ra: float
    dec: float
    resolver: str


class ResolveGetSchema(BaseSchema):
    """Schema defines required parameters for a GET"""

    name: str


class JobSchema(BaseSchema):
    """Full return of Job Information for ACROSSAPIJobs"""

    jobnumber: Optional[int] = None
    reqtype: str
    apiversion: str
    began: datetime
    created: datetime
    expires: datetime
    params: str
    result: Optional[str] = None
    status: Optional[str] = None


class UserArgSchema(BaseSchema):
    username: Optional[str] = "anonymous"
    api_token: Optional[str] = None

    @model_validator(mode="after")
    def username_requires_api_token(self) -> "UserArgSchema":
        if self.username != "anonymous" and self.api_token is None:
            raise ValueError("api_token required if username is set")
        return self


class ACROSSAPIJobsGetSchema(BaseSchema):
    username: str
    begin: Optional[datetime]
    end: Optional[datetime]
    unexpired_only: bool = True
    reqtype: str



import re
import warnings
from datetime import date, datetime, timedelta, timezone
from typing import Optional, Union

import astropy.units as u  # type: ignore
import numpy as np
from astropy.time import Time, TimeDelta  # type: ignore
from astropy.units import Quantity
from dateutil import parser


def tablefy(table: list, header: Optional[list] = None) -> str:
    """Simple HTML table generator

    Parameters
    ----------
    table : list
        Data for table
    header : list, optional
        Headers for table, by default None

    Returns
    -------
    str
        HTML formatted table.
    """

    tab = "<table>"
    if header is not None:
        tab += "<thead>"
        tab += "".join(
            [f"<th style='text-align: left;'>{head}</th>" for head in header]
        )
        tab += "</thead>"

    for row in table:
        tab += "<tr>"
        # Replace any carriage returns with <br>
        row = [f"{col}".replace("\n", "<br>") for col in row]
        tab += "".join([f"<td style='text-align: left;'>{col}</td>" for col in row])
        tab += "</tr>"
    tab += "</table>"
    return tab


# Regex for matching date, time and datetime strings
DATE_REGEX = r"^[0-2]\d{3}-(0?[1-9]|1[012])-([0][1-9]|[1-2][0-9]|3[0-1])?$"
ISO8601_REGEX = r"^([\+-]?\d{4}(?!\d{2}\b))((-?)((0[1-9]|1[0-2])(\3([12]\d|0[1-9]|3[01]))?|W([0-4]\d|5[0-2])(-?[1-7])?|(00[1-9]|0[1-9]\d|[12]\d{2}|3([0-5]\d|6[1-6])))([T\s]((([01]\d|2[0-3])((:?)[0-5]\d)?|24\:?00)([\.,]\d+(?!:))?)?(\17[0-5]\d([\.,]\d+)?)?([zZ]|([\+-])([01]\d|2[0-3]):?([0-5]\d)?)?)?)?$"
DATETIME_REGEX = r"^[0-2]\d{3}-(0?[1-9]|1[012])-([0][1-9]|[1-2][0-9]|3[0-1])[\sT]([0-9]:|[0-1][0-9]:|2[0-3]:)[0-5][0-9]:[0-5][0-9]+(\.\d+)?$"


def convert_timedelta(
    length: Union[str, float, timedelta, TimeDelta, Quantity, None], units=u.day
) -> timedelta:
    """Convert various timedelta formats to swiftdatetime or datetime

    Parameters
    ----------
    length : Union[str, float, timedelta, TimeDelta, Quantity, None]
        Value to be converted.
    units : astropy.units.Quantity, optional
        Unit to use if float/int given. Default is days.

    Returns
    -------
    datetime.timedelta
        Returned timedelta object.

    Raises
    ------
    TypeError
        Raised if incorrect format is given for conversion.
    """

    if units == u.day:
        divisor = 86400.0
    else:
        divisor = 1.0
    if type(length) is Quantity:
        length = length.to(u.day).value
    elif type(length) is timedelta:
        length = length.total_seconds() / divisor
    elif type(length) is TimeDelta:
        length = length.to_datetime().total_seconds() / divisor  # type: ignore
    else:
        try:
            length = float(length)  # type: ignore
        except ValueError:
            raise TypeError(
                f"Length of time should be given as a datetime.timedelta, astropy TimeDelta, astropy quantity or as a number of {units}"
            )
    return timedelta(days=length)  # type: ignore


def convert_to_dt(value: Union[str, date, datetime, Time]) -> Optional[datetime]:
    """Convert various date formats to datetime

    Parameters
    ----------
    value : Union[str, date, datetime, Time]
        Value to be converted.

    Returns
    -------
    datetime.datetime
        Returned datetime object.

    Raises
    ------
    TypeError
        Raised if incorrect format is given for conversion.
    """
    if value is None:
        return None

    if isinstance(value, (str, np.str_)):
        if re.match(DATETIME_REGEX, value):
            # Remove the rogue T in 2023-10-17T00:00:00 style strings
            value = value.replace("T", " ")
            # Figure out if we have decimal places
            if "." in value:
                # Do this because "fromisoformat" is restricted to 0, 3 or 6
                # decimal plaaces
                dtvalue = datetime.strptime(value, "%Y-%m-%d %H:%M:%S.%f")
            else:
                dtvalue = datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        elif re.match(DATE_REGEX, value):
            dtvalue = datetime.strptime(f"{value} 00:00:00", "%Y-%m-%d %H:%M:%S")
        elif re.match(ISO8601_REGEX, value):
            dtvalue = parser.parse(value)
            if dtvalue.tzinfo is None:
                warnings.warn(
                    "ISO8601 formatted dates should be supplied with timezone. ISO8601 dates with no timezone will be assumed to be localtime and then converted to UTC."
                )
            dtvalue = dtvalue.astimezone(timezone.utc).replace(tzinfo=None)
        else:
            raise ValueError(
                "Date/time given as string should 'YYYY-MM-DD HH:MM:SS' or ISO8601 format."
            )
    elif type(value) is date:
        dtvalue = datetime.strptime(f"{value} 00:00:00", "%Y-%m-%d %H:%M:%S")
    elif type(value) is datetime:
        if value.tzinfo is not None:
            # Strip out timezone info and convert to UTC
            value = value.astimezone(timezone.utc).replace(tzinfo=None)
        dtvalue = value  # Just pass through un molested
    elif type(value) is Time:
        dtvalue = value.datetime
    else:
        raise TypeError(
            'DateTime should be given as a datetime, astropy Time, or as string of format "YYYY-MM-DD HH:MM:SS"'
        )

    return dtvalue



from typing import Optional, Union

from astropy.coordinates import SkyCoord  # type: ignore
from astropy.coordinates import Latitude, Longitude  # type: ignore
from astropy.units import Quantity, deg  # type: ignore


def coord_convert(
    coord: Union[float, int, str, Quantity, Longitude, None],
) -> Optional[float]:
    """Convert coordinates of various types either string, integer,
    astropy Longitude or astropy "deg" unit, to a float.

    Parameters
    ----------
    coord : Union[float, str, int, Quantity, Longitude, None]
        Coordinate in one of the types

    Returns
    -------
    float | None
        Coordinate expressed as a float. Just pass through a None.
    """
    if coord is None:
        return None
    if type(coord) is Quantity:
        return coord.to(deg).value
    if type(coord) is Longitude or type(coord) is Latitude:
        return coord.value
    # Universal translator
    return float(coord)


class ACROSSSkyCoord:
    ra: float
    dec: float

    @property
    def skycoord(self) -> SkyCoord:
        """Returns a string representation of the skycoord."""
        if self.ra is None or self.dec is None:
            return None
        else:
            return SkyCoord(self.ra, self.dec, unit="deg")

    @skycoord.setter
    def skycoord(self, coord: SkyCoord):
        """Sets the ra and dec from a SkyCoord."""
        self.ra = coord.icrs.ra.deg
        self.dec = coord.icrs.dec.deg



from dataclasses import dataclass
from typing import Optional


@dataclass
class ACROSSUser:
    """
    Represents an ACROSS user.

    Attributes
    ----------
    username : Optional[str]
        The username of the user.
    api_token : Optional[str]
        The API key associated with the user.

    """

    username: Optional[str] = None
    api_token: Optional[str] = None



from datetime import datetime
from enum import Enum
from io import IOBase
from typing import List, Optional

from pydantic import ConfigDict  # type: ignore

from ..base.schema import BaseSchema, OptionalDateRangeSchema, OptionalPositionSchema


class TOOReason(str, Enum):
    """
    Reasons for rejecting TOO observations

    Attributes
    ----------
    saa
        In SAA
    earth_occult
        Earth occulted
    moon_occult
        Moon occulted
    sun_occult
        Sun occulted
    too_old
        Too old
    other
        Other
    none
        None
    """

    saa = "In SAA"
    earth_occult = "Earth occulted"
    moon_occult = "Moon occulted"
    sun_occult = "Sun occulted"
    too_old = "Too old"
    other = "Other"
    none = "None"


class TOOStatus(str, Enum):
    """
    Enumeration class representing the status of a Target of Opportunity (TOO) request.

    Attributes:
    requested
        The TOO request has been submitted.
    rejected
        The TOO request has been rejected.
    declined
        The TOO request has been declined.
    approved
        The TOO request has been approved.
    executed
        The TOO request has been executed.
    other
        The TOO request has a status other than the predefined ones.
    """

    requested = "Requested"
    rejected = "Rejected"
    declined = "Declined"
    approved = "Approved"
    executed = "Executed"
    deleted = "Deleted"
    other = "Other"


class BurstCubeTriggerInfo(BaseSchema):
    """
    Metadata schema for the BurstCube Target of Opportunity (TOO) request. Note
    that this schema is not strictly defined, keys are only suggested, and
    additional keys can be added as needed.
    """

    trigger_name: Optional[str] = None
    trigger_mission: Optional[str] = None
    trigger_instrument: Optional[str] = None
    trigger_id: Optional[str] = None
    trigger_duration: Optional[float] = None
    classification: Optional[str] = None
    justification: Optional[str] = None

    model_config = ConfigDict(extra="allow")


class BurstCubeTOOSchema(OptionalPositionSchema):
    """
    Schema describing a BurstCube TOO Request.
    """

    id: Optional[int] = None
    created_by: str
    created_on: datetime
    modified_by: Optional[str] = None
    modified_on: Optional[datetime] = None
    trigger_time: datetime
    trigger_info: BurstCubeTriggerInfo
    exposure: int
    offset: int
    reject_reason: TOOReason = TOOReason.none
    status: TOOStatus = TOOStatus.requested
    too_info: str = ""
    healpix_filename: Optional[str] = None
    version: int


class BurstCubeTOODelSchema(BaseSchema):
    """
    Schema for BurstCubeTOO DELETE API call.

    Attributes
    ----------
    id
        The ID of the BurstCubeTOODel object.
    """

    id: int


class BurstCubeTOOPostSchema(OptionalPositionSchema):
    """
    Schema to submit a TOO request for BurstCube.
    """

    trigger_time: datetime
    trigger_info: BurstCubeTriggerInfo
    exposure: int = 200
    offset: int = -50
    healpix_file: Optional[IOBase] = None


class BurstCubeTOOGetSchema(BaseSchema):
    """
    Schema for BurstCubeTOO GET request.
    """

    id: int


class BurstCubeTOOPutSchema(BurstCubeTOOPostSchema):
    """
    Schema for BurstCubeTOO PUT request.
    """

    id: int
    trigger_time: datetime
    trigger_info: BurstCubeTriggerInfo
    exposure: int = 200
    offset: int = -50
    status: TOOStatus
    healpix_file: Optional[IOBase] = None


class BurstCubeTOORequestsGetSchema(OptionalDateRangeSchema):
    """
    Schema for GET requests to retrieve BurstCube Target of Opportunity (TOO) requests.
    """

    duration: Optional[float] = None
    limit: Optional[int] = None
    offset: Optional[int] = None


class BurstCubeTOORequestsSchema(BaseSchema):
    """
    Schema for BurstCube TOO requests.
    """

    entries: List[BurstCubeTOOSchema]



import io
from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Union

from pydantic import FilePath

from ..across.resolve import ACROSSResolveName
from ..base.common import ACROSSBase
from .constants import MISSION
from .schema import (
    BurstCubeTOOGetSchema,
    BurstCubeTOOPostSchema,
    BurstCubeTOOPutSchema,
    BurstCubeTOORequestsGetSchema,
    BurstCubeTOORequestsSchema,
    BurstCubeTOOSchema,
    BurstCubeTriggerInfo,
    TOOReason,
    TOOStatus,
)


@dataclass
class TOO(ACROSSBase, ACROSSResolveName):
    """
    Class representing a Target of Opportunity (TOO) request.

    Parameters:
    ----------
    api_token: str
        The api_token for login (optional).
    trigger_time : datetime
        The time of the trigger.
    ra : Optional[float]
        The right ascension of the target (optional).
    dec : Optional[float]
        The declination of the target (optional).
    begin : datetime
        The start time of the TOO observation.
    end : datetime
        The end time of the TOO observation.
    exposure : float
        The exposure time for the TOO observation.
    offset : float
        The offset for the TOO observation.
    healpix_filename : Optional[FilePath]
        The healpix filename that represents the object localization for the TOO. This should be a file
        on disk.
    healpix_file : Union[io.BytesIO, io.BufferedReader, None]
        The healpix file handle for the TOO observation, takes a file like object.
    trigger_info : BurstCubeTriggerInfo
        The trigger information for the TOO observation.
    too_info : str
        The information about the TOO observation.
    status : str
        The status of the TOO request.
    id : id
        The ID of the TOO request.


    Attributes:
    ----------
    created_on : datetime
        The time at which the TOO request was made.
    created_by : str
        The user who made the TOO request.
    modified_on : datetime
        The time at which the TOO request was last modified.
    modified_by : str
        The user who modified the TOO request.
    reject_reason : str
        The reason for the TOO request being rejected.
    status : str
        The status of the TOO request.
    too_info : str
        The information about the TOO request.
    id : str
        The ID of the TOO request.
    """

    api_token: Optional[str] = None
    id: Optional[int] = None
    trigger_time: Optional[datetime] = None
    created_by: Optional[str] = None
    created_on: Optional[datetime] = None
    modified_by: Optional[str] = None
    modified_on: Optional[datetime] = None
    trigger_info: BurstCubeTriggerInfo = field(default_factory=BurstCubeTriggerInfo)
    exposure: float = 200
    offset: float = -50
    too_info: str = ""
    ra: Optional[float] = None
    dec: Optional[float] = None
    error_radius: Optional[float] = None
    healpix_filename: Optional[FilePath] = None
    healpix_file: Union[io.BytesIO, io.BufferedReader, None] = None
    reject_reason: TOOReason = TOOReason.none
    status: TOOStatus = TOOStatus.requested
    version: Optional[int] = None

    # API definitions
    _mission = MISSION
    _api_name = "TOO"
    _schema = BurstCubeTOOSchema
    _put_schema = BurstCubeTOOPutSchema
    _post_schema = BurstCubeTOOPostSchema
    _get_schema = BurstCubeTOOGetSchema
    _del_schema = BurstCubeTOOGetSchema

    @classmethod
    def submit_too(cls, **kwargs):
        """
        Submit a TOO request.
        """
        for k, a in kwargs.items():
            if k in cls._post_schema.model_fields.keys():
                setattr(cls, k, a)

        if cls.validate_post():
            cls.post()

    @property
    def _table(self):
        return (
            [
                "TOO ID",
                "Submitted",
                "Submitter",
                "Trigger Time",
                "Mission",
                "Instrument",
                "ID",
                "Status",
                "Reason",
            ],
            [
                [
                    self.id,
                    self.created_on,
                    self.created_by,
                    self.trigger_time,
                    self.trigger_info.trigger_mission,
                    self.trigger_info.trigger_instrument,
                    self.trigger_info.trigger_id,
                    self.status.value,
                    self.reject_reason.value,
                ]
            ],
        )


class TOORequests(ACROSSBase):
    """
    Represents a Targer of Opportunity (TOO) request.

    Attributes
    ----------
    begin : datetime
        The start time of the observation.
    end : datetime
        The end time of the observation.
    limit : int
        The maximum number of entries for the observation.
    trigger_time : datetime
        The time at which the observation should be triggered.
    entries : list
        The list of entries for the observation.
    """

    _mission = MISSION
    _api_name = "TOO"
    _schema = BurstCubeTOORequestsSchema
    _get_schema = BurstCubeTOORequestsGetSchema

    def __init__(self, **kwargs):
        self.entries = []
        self.begin = None
        self.end = None
        self.limit = None
        for k, a in kwargs.items():
            if k in self._get_schema.model_fields.keys():
                setattr(self, k, a)

        # As this is a GET only class, we can validate and get the data
        if self.validate_get():
            self.get()

        # Convert the entries to a list of TOO objects
        self.entries = [TOO(**entry.__dict__) for entry in self.entries]

    def by_id(self, id):
        """
        Get a TOO request by ID.
        """
        for entry in self.entries:
            if entry.id == id:
                return entry

    @property
    def _table(self):
        return (
            [
                "TOO ID",
                "Submitted",
                "Submitter",
                "Trigger Time",
                "Mission",
                "Instrument",
                "ID",
                "Status",
                "Reason",
            ],
            [
                [
                    entry.id,
                    entry.created_on,
                    entry.created_by,
                    entry.trigger_time,
                    entry.trigger_info.trigger_mission,
                    entry.trigger_info.trigger_instrument,
                    entry.trigger_info.trigger_id,
                    entry.status.value,
                    entry.reject_reason.value,
                ]
                for entry in self.entries
            ],
        )


# Alias
BurstCubeTOO = TOO
BurstCubeTOORequests = TOORequests
submit_too = TOO.submit_too
burstcube_submit_too = TOO.submit_too



from .toorequest import (
    BurstCubeTOORequests,  # noqa:F401
    BurstCubeTOO,  # noqa:F401
    TOORequests,  # noqa:F401
    TOO,  # noqa:F401
)
from .schema import BurstCubeTriggerInfo  # noqa:F401

TriggerInfo = BurstCubeTriggerInfo  # noqa:F401



from datetime import datetime
from typing import Any, Optional

from astropy.coordinates import SkyCoord  # type: ignore
from pydantic import BaseModel, ConfigDict, Field, model_validator

from ..functions import convert_to_dt  # type: ignore
from .coords import coord_convert  # type: ignore


class BaseSchema(BaseModel):
    """Base schema for all other schemas"""

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)

    @property
    def _table(self):
        """Get the table representation of the schema"""
        header = self.model_fields.keys()
        return list(header), [list(self.model_dump().values())]


class CoordSchema(BaseSchema):
    """Schema that defines basic RA/Dec"""

    ra: float = Field(ge=0, lt=360)
    dec: float = Field(ge=-90, le=90)

    @model_validator(mode="before")
    @classmethod
    def convert_coord(cls, data: Any) -> Any:
        """Convert the coordinate data to a specific format"""
        if isinstance(data, dict):
            for key in data.keys():
                if key == "ra" or key == "dec":
                    data[key] = coord_convert(data[key])
        else:
            data.ra = coord_convert(data.ra)
            data.dec = coord_convert(data.dec)
        return data

    @property
    def skycoord(self) -> SkyCoord:
        """Get the SkyCoord representation of the coordinates"""
        return SkyCoord(self.ra, self.dec, unit="deg")


class OptionalCoordSchema(BaseSchema):
    """Schema that defines optional RA/Dec"""

    ra: Optional[float] = Field(ge=0, lt=360, default=None)
    dec: Optional[float] = Field(ge=-90, le=90, default=None)

    @model_validator(mode="before")
    @classmethod
    def coord_convert(cls, data: Any) -> Any:
        """Convert the coordinate data to a specific format"""
        if isinstance(data, dict):
            for key in data.keys():
                if key == "ra" or key == "dec":
                    data[key] = coord_convert(data[key])
        elif hasattr(data, "ra") and hasattr(data, "dec"):
            data.ra = coord_convert(data.ra)
            data.dec = coord_convert(data.dec)
        return data

    @model_validator(mode="after")
    @classmethod
    def check_ra_dec(cls, data: Any) -> Any:
        """Check if RA and Dec are both set or both not set"""
        if data.ra is None or data.dec is None:
            assert data.ra == data.dec, "RA/Dec should both be set, or both not set"
        return data

    @property
    def skycoord(self) -> Optional[SkyCoord]:
        """Get the SkyCoord representation of the coordinates"""
        if self.ra is not None and self.dec is not None:
            return SkyCoord(self.ra, self.dec, unit="deg")
        return None


class OptionalPositionSchema(OptionalCoordSchema):
    """Schema that defines position with error"""

    error_radius: Optional[float] = None


class DateRangeSchema(BaseSchema):
    """Schema that defines date range"""

    begin: datetime
    end: datetime

    @model_validator(mode="before")
    @classmethod
    def convert_date(cls, data: Any) -> datetime:
        """Check if the begin and end dates are both set or both not set"""
        if isinstance(data, dict):
            for key in data.keys():
                if key == "begin" or key == "end":
                    data[key] = convert_to_dt(data[key])
        else:
            data.begin = convert_to_dt(data.begin)
            data.end = convert_to_dt(data.end)
        return data

    @model_validator(mode="after")
    @classmethod
    def check_dates(cls, data: Any) -> Any:
        """Check if the begin and end dates are both set or both not set"""
        assert data.begin <= data.end, "End date should not be before begin."

        return data


class OptionalDateRangeSchema(BaseSchema):
    """Schema that defines optional date range"""

    begin: Optional[datetime] = None
    end: Optional[datetime] = None

    @model_validator(mode="before")
    @classmethod
    def convert_date(cls, data: Any) -> datetime:
        """Check if the begin and end dates are both set or both not set"""
        if isinstance(data, dict):
            for key in data.keys():
                if key == "begin" or key == "end":
                    data[key] = convert_to_dt(data[key])
        else:
            data.begin = convert_to_dt(data.begin)
            data.end = convert_to_dt(data.end)
        return data

    @model_validator(mode="after")
    @classmethod
    def check_dates(cls, data: Any) -> Any:
        """Check if the begin and end dates are both set or both not set"""
        if data.begin is None or data.end is None:
            assert (
                data.begin == data.end
            ), "Begin/End should both be set, or both not set."
        if data.begin is not None and data.end is not None:
            assert data.begin <= data.end, "End date should not be before begin."

        return data



import json
import warnings
from pathlib import PosixPath
from typing import Any, Dict, Optional, Tuple, Type

import requests

from ..constants import API_URL
from ..functions import tablefy
from .schema import BaseSchema


class ACROSSBase:
    """
    Base class for ACROSS API Classes including common methods for all API classes.
    """

    # Type hints
    entries: list

    # API descriptors type hints
    _schema: Type[BaseSchema]
    _get_schema: Type[BaseSchema]
    _put_schema: Type[BaseSchema]
    _post_schema: Type[BaseSchema]
    _del_schema: Type[BaseSchema]

    _mission: str
    _api_name: str = __name__

    def __getitem__(self, i):
        return self.entries[i]

    def api_url(self, argdict) -> str:
        """
        URL for this API call.

        Returns
        -------
        str
            URL for API call
        """
        # If arguments has `id` in it, then put this in the path
        if "id" in argdict.keys() and argdict["id"] is not None:
            return f"{API_URL}{self._mission.lower()}/{self._api_name.lower()}/{argdict['id']}"
        return f"{API_URL}{self._mission.lower()}/{self._api_name.lower()}"

    @property
    def schema(self) -> Any:
        """Return pydantic schema for this API class

        Returns
        -------
        object
            Pydantic Schema
        """
        return self._schema.model_validate(self)

    @property
    def parameters(self) -> dict:
        """
        Return parameters as dict

        Returns
        -------
        dict
            Dictionary of parameters
        """
        return {k: v for k, v in self._schema.model_validate(self) if v is not None}

    @parameters.setter
    def parameters(self, params: dict):
        """
        Set API parameters from a given dict which is validated from self._schema

        Parameters
        ----------
        params : dict
            Dictionary of class parameters
        """
        for k, v in self._schema(**params):
            if hasattr(self, k) and v is not None:
                setattr(self, k, v)

    @property
    def headers(self) -> dict:
        # If we have a api_token, then use this to authenticate
        if hasattr(self, "api_token"):
            headers = {"Authorization": f"Bearer {self.api_token}"}
        else:
            headers = {}
        return headers

    def get(self) -> bool:
        """
        Perform a 'GET' submission to ACROSS API. Used for fetching
        information.

        Returns
        -------
        bool
            Was the get successful?

        Raises
        ------
        HTTPError
            Raised if GET doesn't return a 200 response.
        """
        if self.validate_get():
            # Create an array of parameters from the schema
            get_params = {
                key: value for key, value in self._get_schema.model_validate(self)
            }
            # Do the GET request
            req = requests.get(
                self.api_url(get_params),
                params=get_params,
                headers=self.headers,
                timeout=60,
            )
            if req.status_code == 200:
                # Parse, validate and record values from returned API JSON
                for k, v in self._schema.model_validate(req.json()):
                    setattr(self, k, v)
                return True
            elif req.status_code == 404:
                """Handle 404 errors gracefully, by issuing a warning"""
                warnings.warn(req.json()["detail"])
            else:
                # Raise an exception if the HTML response was not 200
                req.raise_for_status()
        return False

    def delete(self) -> bool:
        """
        Perform a 'GET' submission to ACROSS API. Used for fetching
        information.

        Returns
        -------
        bool
            Was the get successful?

        Raises
        ------
        HTTPError
            Raised if GET doesn't return a 200 response.
        """
        if self.validate_del():
            # Create an array of parameters from the schema
            del_params = {
                key: value for key, value in self._del_schema.model_validate(self)
            }
            # Do the DELETE request
            req = requests.delete(
                self.api_url(del_params),
                params=del_params,
                headers=self.headers,
                timeout=60,
            )
            if req.status_code == 200:
                # Parse, validate and record values from returned API JSON
                for k, v in self._schema.model_validate(req.json()):
                    setattr(self, k, v)
                return True
            else:
                # Raise an exception if the HTML response was not 200
                req.raise_for_status()
        return False

    def put(self, payload={}) -> bool:
        """
        Perform a 'PUT' submission to ACROSS API. Used for pushing/replacing
        information.

        Returns
        -------
        bool
            Was the get successful?

        Raises
        ------
        HTTPError
            Raised if PUT doesn't return a 201 response.
        """

        if self.validate_put():
            # Extract any files out of the arguments
            files: Dict[str, Tuple[Optional[str], Any, str]] = {
                key.replace("filename", "file"): (
                    # Return either the existing filelike object, or open the file
                    value.name,
                    (
                        getattr(self, key.replace("filename", "file"))
                        if hasattr(self, key.replace("filename", "file"))
                        else value.open("rb")
                    ),
                    "application/octet-stream",
                )
                for key, value in self._put_schema.model_validate(self)
                if type(value) is PosixPath
            }

            # Extract all POST parameters from the schema
            all_params = self._put_schema.model_validate(self).model_dump(
                mode="json", exclude_none=True
            )

            # Extract query parameters
            put_params = {
                k: all_params[k]
                for k in all_params
                if not isinstance(all_params[k], dict)
            }

            # URL for this API call
            api_url = self.api_url(put_params)
            if "id" in put_params.keys():
                put_params.pop("id")  # Remove id from query parameters

            # Add any json data to files
            for k in all_params:
                if isinstance(all_params[k], dict):
                    files[k] = (None, json.dumps(all_params[k]), "application/json")

            # Submit request
            req = requests.put(
                api_url,
                params=put_params,
                headers=self.headers,
                files=files,
                timeout=60,
            )

            if req.status_code == 201:
                # Parse, validate and record values from returned API JSON
                for k, v in self._schema.model_validate(req.json()):
                    setattr(self, k, v)
                return True
            else:
                print("ERROR: ", req.status_code, req.json())
                req.raise_for_status()
        return False

    def post(self) -> bool:
        """
        Perform a 'POST' submission to ACROSS API. For submitting new
        information to the API.

        Returns
        -------
        bool
            Was the get successful?

        Raises
        ------
        HTTPError
            Raised if POST doesn't return a 201 response.
        """
        if self.validate_post():
            # Extract any files out of the arguments
            files: Dict[str, Tuple[Optional[str], Any, str]] = {
                key: (
                    PosixPath(getattr(self, key + "name")).name
                    if getattr(self, key) is not None
                    else "healpix_file.fits",
                    open(getattr(self, key + "name"), "rb"),
                    "application/octet-stream",
                )
                if getattr(self, key) is None
                and getattr(self, key + "name") is not None
                else (
                    PosixPath(getattr(self, key).name).name
                    if getattr(self, key) is not None
                    else None,
                    getattr(self, key),
                    "application/octet-stream",
                )
                for key, value in self._post_schema.model_validate(self)
                if "_file" in key
            }

            # If files are gzipped, then set the content type to
            # application/x-gzip
            for key in files:
                filename = files[key][0]
                if isinstance(filename, str):
                    if ".gz" in filename:
                        files[key] = (
                            files[key][0],
                            files[key][1],
                            "application/x-gzip",
                        )

            # Extract all POST parameters from the schema
            all_params = self._post_schema.model_validate(self).model_dump(
                mode="json", exclude_none=True
            )

            # Extract query parameters
            post_params = {
                k: all_params[k]
                for k in all_params
                if not isinstance(all_params[k], dict)
            }

            # Add any json data to files
            for k in all_params:
                if isinstance(all_params[k], dict):
                    files[k] = (None, json.dumps(all_params[k]), "application/json")
            # Submit request
            req = requests.post(
                self.api_url(post_params),
                params=post_params,
                headers=self.headers,
                files=files,
                timeout=60,
            )

            if req.status_code == 201:
                # Parse, validate and record values from returned API JSON
                print(req.json())
                for k, v in self._schema.model_validate(req.json()):
                    setattr(self, k, v)
                return True
            elif req.status_code == 200:
                warnings.warn(req.json()["detail"])
                return False
            else:
                # Raise an exception if the HTML response was not 200
                try:
                    req.raise_for_status()
                except requests.HTTPError as exc:
                    print(f"HTTP Exception: {exc}: {req.json()['detail']}.")

        return False

    def validate_get(self) -> bool:
        """Validate arguments for GET

        Returns
        -------
        bool
            Do arguments validate? True | False

        Raises
        ------
        ValidationError
            If arguments don't validate
        """
        if hasattr(self, "_get_schema"):
            self._get_schema.model_validate(self)
        else:
            warnings.warn("GET not allowed for this class.")
            return False
        return True

    def validate_put(self) -> bool:
        """Validate if value to be PUT matches Schema

        Returns
        -------
        bool
            Is it validated? True | False

        Raises
        ------
        ValidationError
            If the value to be PUT doesn't match the Schema


        """
        if hasattr(self, "_put_schema"):
            self._put_schema.model_validate(self.__dict__)
        else:
            warnings.warn("PUT not allowed for this class.")
            return False
        return True

    def validate_post(self) -> bool:
        """Validate if value to be POST matches Schema

        Returns
        -------
        bool
            Is it validated? True | False
        """
        if hasattr(self, "_post_schema"):
            self._post_schema.model_validate(self.__dict__)
        else:
            warnings.warn("POST not allowed for this class.")
            return False
        return True

    def validate_del(self) -> bool:
        """Validate if value to be POST matches Schema

        Returns
        -------
        bool
            Is it validated? True | False
        """
        if hasattr(self, "_del_schema"):
            self._del_schema.model_validate(self.__dict__)
        else:
            warnings.warn("DELETE not allowed for this class.")
            return False
        return True

    @property
    def _table(self) -> tuple:
        """
        Table with head showing results of the API query.

        Returns
        -------
        tuple
            Tuple containing two lists, the header and the table data
        """
        if hasattr(self, "entries") and len(self.entries) > 0:
            header = self.entries[0]._table[0]
            table = [t._table[1][0] for t in self.entries]
        else:
            # Start with arguments
            if hasattr(self, "_get_schema"):
                _parameters = list(self.parameters.keys())
            else:
                _parameters = []
            _parameters += list(self.parameters.keys())
            # Don't include client_id/client_secret in table
            try:
                _parameters.pop(_parameters.index("client_id"))
                _parameters.pop(_parameters.index("client_secret"))
            except ValueError:
                pass

            # Removed repeated values
            _parameters = list(set(_parameters))

            header = [par for par in _parameters]
            table = []
            for row in _parameters:
                value = getattr(self, row)
                table.append(value)
            table = [table]
        return header, table

    def _repr_html_(self) -> str:
        """Return a HTML summary of the API data, for e.g. Jupyter.

        Returns
        -------
        str
            HTML summary of data
        """
        header, table = self._table
        if len(table) > 0:
            return tablefy(table, header)
        else:
            return "No data"

    def __repr__(self) -> str:
        # print a string showing the API call and arguments with their values
        # in a way that can be copied and pasted into a script
        args = ",".join([f"{k}={v}" for k, v in self.parameters.items()])
        return f"{self.__class__.__name__}({args})"



from datetime import datetime
from typing import Optional


class ACROSSDateRange:
    """
    Represents a date range in the ACROSS API.

    Parameters
    ----------
    begin : datetime
        The start date of the range.
    end : Optional[datetime], optional
        The end date of the range.

    Attributes
    ----------
    begin : datetime
        The start date of the range.
    end : Optional[datetime]
        The end date of the range.
    _length : Union[float, timedelta, str, None]
        The length of the date range.

    Properties
    ----------
    length : Union[float, timedelta, str, None]
        The length of the date range.

    Methods
    -------
    None
    """

    begin: datetime
    end: Optional[datetime]
